package misc;

import java.math.BigDecimal;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.Locale;

import fr.cnes.genius.exception.GFileManipulatorException;
import fr.cnes.genius.main.GFileManipulation;
import fr.cnes.genopus.exception.GPOrbitException;
import fr.cnes.genopus.orbits.GPOrbit;
import fr.cnes.sirius.addons.patriusdataset.PatriusDataset;
import fr.cnes.sirius.patrius.bodies.CelestialBody;
import fr.cnes.sirius.patrius.bodies.CelestialBodyFactory;
import fr.cnes.sirius.patrius.frames.Frame;
import fr.cnes.sirius.patrius.frames.FramesFactory;
import fr.cnes.sirius.patrius.frames.H0MinusNFrame;
import fr.cnes.sirius.patrius.frames.TranslatedFrame;
import fr.cnes.sirius.patrius.frames.configuration.FramesConfiguration;
import fr.cnes.sirius.patrius.frames.configuration.FramesConfigurationFactory;
import fr.cnes.sirius.patrius.math.geometry.euclidean.threed.Vector3D;
import fr.cnes.sirius.patrius.math.util.FastMath;
import fr.cnes.sirius.patrius.orbits.ApsisOrbit;
import fr.cnes.sirius.patrius.orbits.CartesianOrbit;
import fr.cnes.sirius.patrius.orbits.Orbit;
import fr.cnes.sirius.patrius.orbits.PositionAngle;
import fr.cnes.sirius.patrius.orbits.pvcoordinates.PVCoordinates;
import fr.cnes.sirius.patrius.time.AbsoluteDate;
import fr.cnes.sirius.patrius.time.TimeScale;
import fr.cnes.sirius.patrius.time.TimeScalesFactory;
import fr.cnes.sirius.patrius.utils.exception.PatriusException;

/**
 * Tutorial for using GPOrbit.
 * @author goesterjf
 *
 */
public class TestGPOrbitPrecision {

    /**
     * @param args
     * @throws OrekitException 
     * @throws GPOrbitException 
     * @throws GFileManipulatorException 
     */
    public static void main(String[] args) throws PatriusException, GPOrbitException, GFileManipulatorException {
        
        Locale.setDefault(Locale.US);

        PatriusDataset.addResourcesFromPatriusDataset();
        
        FramesConfiguration conf = FramesConfigurationFactory.getSimpleConfiguration(true);
        FramesFactory.setConfiguration(conf);
        
        // First case
        
        System.out.println("First case ...");
        System.out.println();
        
        // Orbit data
        final TimeScale UTC = TimeScalesFactory.getUTC();
        AbsoluteDate date = new AbsoluteDate("2020-01-01T00:00:00.000", UTC);
        Frame frame = FramesFactory.getGCRF();
        final double hp = 400.e+3;
        final double ha = 400.e+3;
        final double inc = FastMath.toRadians(51.6);
        final double gom = FastMath.toRadians(0.);
        final double pom = FastMath.toRadians(0.);
        final double ano = FastMath.toRadians(0.);
        final double mu = 398600.4415e+9;
        final double req = 6378.1363e+3;
        Orbit orbit = new ApsisOrbit(hp+req, ha+req, inc, pom, gom, ano, PositionAngle.TRUE, frame, date, mu);
        
        GPOrbit gpPorbit  = new GPOrbit("Orbit");        
        GFileManipulation.readConfig("GPOrbitPrecision.xml", "localTest", gpPorbit, true);        
        Orbit orbitXml = gpPorbit.getPatriusObject();
        
        print(orbit, orbitXml);
        
        // Second case
        
        System.out.println("Second case ...");
        System.out.println();
        
        gpPorbit  = new GPOrbit("Orbit");        
        GFileManipulation.readConfig("GPOrbitPrecision2.xml", "localTest", gpPorbit, true);        
        orbitXml = gpPorbit.getPatriusObject();
        
        date = new AbsoluteDate("2024-01-08T05:29:26.240", TimeScalesFactory.getTAI());
        final CelestialBody moon = CelestialBodyFactory.getMoon();
        final Frame icrf = FramesFactory.getICRF();
        frame = new TranslatedFrame(icrf, moon, "MOON_ICRF");
        Vector3D pos = new Vector3D(3.22263678900463E5, -1.47914373363123E6, 2.89593173838077E6);
        Vector3D vel = new Vector3D(1.50208151682368E3, -6.4938778076607E2, -4.70989850772853E2);
        
        PVCoordinates pv = new PVCoordinates(pos, vel);
        orbit = new CartesianOrbit(pv, frame, date, mu);
//        orbit = new CartesianOrbit(pv, frame, orbitXml.getDate(), mu);
        
        print(orbit, orbitXml);

        // Third case
        
        System.out.println("Third case ...");
        System.out.println();
        
        date = new AbsoluteDate("2022-09-04T01:28:26.000", TimeScalesFactory.getUTC());   
        final AbsoluteDate h0 = new AbsoluteDate("2022-09-04T01:02:36.000", TimeScalesFactory.getUTC());
        frame = FramesFactory.getH0MinusN("H0_MINUS_N", h0, 0., FastMath.toRadians(-5.277E1));
        System.out.println("FastMath:" + FastMath.toRadians(-5.277E1));
        System.out.println("Math:" + Math.toRadians(-5.277E1));
        pos = new Vector3D(-5.047106E5, 7.335901E6, -4.368592E5);
        vel = new Vector3D(-6.299652E3, 2.479892E3, -1.06006E3);
        pv = new PVCoordinates(pos, vel);
        orbit = new CartesianOrbit(pv, frame, date, mu);
        
        gpPorbit  = new GPOrbit("Orbit");        
        GFileManipulation.readConfig("GPOrbitPrecision3Deg.xml", "localTest", gpPorbit, true);        
        orbitXml = gpPorbit.getPatriusObject();
        
        print(orbit, orbitXml);
        
        printDate(((H0MinusNFrame)frame).getH0(), ((H0MinusNFrame)orbitXml.getFrame()).getH0());
        printLine(((H0MinusNFrame)frame).getLongitude(), ((H0MinusNFrame)orbitXml.getFrame()).getLongitude());
        printLine(((H0MinusNFrame)frame).getN(), ((H0MinusNFrame)orbitXml.getFrame()).getN());
        System.out.println();
        
        final Frame eme2000 = FramesFactory.getEME2000();
        orbit = new CartesianOrbit(orbit.getPVCoordinates(eme2000), eme2000, orbit.getDate(), orbit.getMu());
        orbitXml = new CartesianOrbit(orbitXml.getPVCoordinates(eme2000), eme2000, orbitXml.getDate(), orbitXml.getMu());

        print(orbit, orbitXml);
        
//        final MathContext MC = new MathContext(512, RoundingMode.HALF_UP);
        final MathContext MC = new MathContext(17, RoundingMode.HALF_EVEN);
        
        
        final double iniValue = -5.277E1;
        System.out.println("Initial double value:                  " + iniValue);
        final BigDecimal iniValueBig = BigDecimal.valueOf(iniValue);
        System.out.println("Initial big value:                     " + iniValueBig);
        
//        final double convertFactor = 0.017453292519943295;
        final double convertFactor = StrictMath.toRadians(1.);
        final int convertFactorToSIExpBaseTen = 0;
        
        System.out.println();
        System.out.println("Conversion factor (double):            " + convertFactor);
        final BigDecimal factorBig = new BigDecimal(convertFactor, MC);
        System.out.println("Conversion factor (big):               " + factorBig);
        final BigDecimal factorBigTen = factorBig.multiply(BigDecimal.valueOf(StrictMath.pow(10, convertFactorToSIExpBaseTen)));
        System.out.println("Conversion factor with power 10 (big): " + factorBigTen);
        final BigDecimal factorBigTenScaled = factorBig.multiply(BigDecimal.valueOf(StrictMath.pow(10, convertFactorToSIExpBaseTen)), MC);
        System.out.println("Scaled conversion factor with power 10 (big): " + factorBigTenScaled);
        
        System.out.println();
        final BigDecimal valModif = iniValueBig.multiply(factorBigTenScaled, MC);
        System.out.println("Modified big value:                    " + valModif);
        final double value = valModif.doubleValue();
        System.out.println("Modified double value:                 " + value);
        
        System.out.println("Modified value using StrictMath:       " + StrictMath.toRadians(iniValue));
        System.out.println("Modified value using double factor:    " + iniValue*convertFactor);
        
        System.out.println("Modified value using converted factor: " + iniValueBig.doubleValue()*factorBigTenScaled.doubleValue());


        
    }
    
    private static void printDate(final AbsoluteDate date1, final AbsoluteDate date2) {
        
        System.out.println(String.format("%s / %s / %22.15e", date1.toString(), date2.toString(), date2.durationFrom(date1)));
    }
    
    private static void print (final Orbit orbit1, final Orbit orbit2) {
        
        printDate(orbit1.getDate(), orbit2.getDate());

        printLine(orbit1.getPVCoordinates().getPosition().getX(), orbit1.getPVCoordinates().getPosition().getX());
        printLine(orbit1.getPVCoordinates().getPosition().getY(), orbit2.getPVCoordinates().getPosition().getY());
        printLine(orbit1.getPVCoordinates().getPosition().getZ(), orbit2.getPVCoordinates().getPosition().getZ());
        printLine(orbit1.getPVCoordinates().getVelocity().getX(), orbit2.getPVCoordinates().getVelocity().getX());
        printLine(orbit1.getPVCoordinates().getVelocity().getY(), orbit2.getPVCoordinates().getVelocity().getY());
        printLine(orbit1.getPVCoordinates().getVelocity().getZ(), orbit2.getPVCoordinates().getVelocity().getZ());
        printLine(orbit1.getMu(), orbit2.getMu());
        System.out.println();

    }
    
    private static void printLine (final double val1, final double val2) {
        
        System.out.println(String.format("%22.15e / %22.15e / %22.15e",  val1*.001, val2*.001, (val2-val1)));

    }

}
